//////////////////////////////////////////////////////////////////
//
//  NAME: Joon S. Lee (jl)
//  
//  FILE: PCPreCondRule.C
//  PROJ: COKO
//  DATE: 6/96
// 
//  SYNOPSIS
//  This class is for PreCondRule Declaration/Invoke statements.
//
/////////////////////////////////////////////////////////////////
#include "PCPreCondRule.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CPreCondRule (CPattern* cp1, CPattern* cp2)
// Parameters: (CPattern*, CPattern*)
//
//  Constructor for preconditioned rule. 
//  Takes three arguments that are precondition list, patterns of
//  lhs and rhs.
//  This class is a subclass of CRule class. 
//         								    
//////////////////////////////////////////////////////////////////////////////

CPreCondRule::CPreCondRule (CPreCond* pc, CPattern* cp1, CPattern* cp2) : CRule(cp1,cp2)
{
  _precond = pc;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CPreCondRule 
// Parameters: (void)
//
// Destructor
//         								    
//////////////////////////////////////////////////////////////////////////////

CPreCondRule::~CPreCondRule(void)
{  
  if(_precond) {delete _precond; _precond = NULL; }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters: (CState*)
//
//  Executes the given statement.
//  Takes the state of the program before the statement
//  is executed, and then returns the state of the 
//  program after the statement is executed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CPreCondRule::Exec(CState* s)
{

  if (_debug_)
	cout << "-- CPreCondRule Begins" << endl;

  s = _lhs->match(s);

  if (s->ISuccess()) { 
	// cout << "matching successfully ended" << endl; 
	s = _precond->Exec(s, (s->Store())->Copy_kTree());
  };

  if (s->ISuccess()) { 
	// cout << "processing prolog successfully ended" << endl; 
	s = _rhs->build(s);
  };

  if (_debug_)
	cout << "-- CPreCondRule Ends" << endl;

  return s;
}

