//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Joon S. Lee (jl), Kee-Eung Kim (kek)  
//  FILE: PCRuleInvokeStmt.C
//  PROJ: COKO
//  DATE: 5/5/97
//
// SYNOPSIS
//  Method definitions for CRuleInvokeStmt class.
//  
//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include "PCStmt.H"
#include "PCState.H"
#include "PCRuleInvokeStmt.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CRuleInvokeStmt
// Parameters: (CRule*, char*)
//
// Constructor for rule invoke.
//         								    
//////////////////////////////////////////////////////////////////////////////

CRuleInvokeStmt::CRuleInvokeStmt(CRule *rule, char *iname):
  _first_r(rule),
  _first_rb(NULL),
  _second(NULL)
{
  if (iname)
  	{
  	  _second = new char[strlen(iname)+1];
  	  strcpy(_second, iname);
  	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CRuleInvokeStmt
// Parameters: (CRuleBlock*, char*)
//
// Constructor for rule block invoke.
//         								    
//////////////////////////////////////////////////////////////////////////////

CRuleInvokeStmt::CRuleInvokeStmt(CRuleBlock *rule_block, char *iname):
  _first_rb(rule_block),
  _first_r(NULL),
  _second(NULL)
{
  if (iname)
  	{
  	  _second = new char[strlen(iname)+1];
  	  strcpy(_second, iname);
  	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CRuleInvokeStmt
// Parameters: (void)
//
// Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CRuleInvokeStmt::~CRuleInvokeStmt(void)
{
  _first_rb = NULL;
  _first_r = NULL;
  if (_second) { delete [] _second; _second = NULL; }

  // We do not free _first_rb nor _first_r. 
  // The rule will be deleted outside CStmts explicitly. Don't worry about it!
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters: (CState*)
//
// What to do when executed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState *CRuleInvokeStmt::Exec(CState *s)
{
  if(_debug_)
	cout << "-- RuleInvokeStmt Begins" << endl;

  NonType * old_root = NULL;  

  if (s->Store())
	old_root = s->Store()->Copy_kTree();
  else
	{
	  cout << "ERROR @ CRuleInvokeStmt : Exec" << endl;
	  cout << "_Store is NULL error" << endl;
	  exit(1);
	}
  
  CIdent *matchNode;
  int is_matchNode_new_one = 0;

  // If we are invoking with an identifier, matchNode will have information
  // about the matching node in the query tree. Else (we are invoking with
  // local root of the query tree), matchnode information will correspond
  // to the local root node.
  
  if (_second) 
	{
	  s->IDeleteInvDep(_second);
	  matchNode = s->IFindDeep(_second);
	}
  else 
	{
	  s->CDeleteInvDep();
	  s->CCleanInvDep();
	  matchNode = new CIdent(s->Store(), 1);
	  is_matchNode_new_one = 1;
	}

  // Make the local root be the matching node.

  s->SetStore(matchNode->Child());

  // Flush out the identifier environment 
  s->INew();

  // Execute!
  if(_first_rb) 
	{
	  DEPList* tmpDEPList = s->GetInvDep();
	  DEPList* emptyDEPList = new DEPList();
	  s->SetInvDep(emptyDEPList);

	  CIEnvStack * tmpStack = s->IEnv();
	  CIEnvStack * emptyStack = new CIEnvStack();
	  s->SetIEnv(emptyStack);
	  //	  s->INew();

	  s = _first_rb->Exec(s);

	  //	  s->IOld();
	  emptyDEPList = NULL;
	  s->DelInvDep();
	  s->SetInvDep(tmpDEPList);
	  s->SetIEnv(tmpStack);
	  if (emptyStack) { delete emptyStack; emptyStack = NULL; }

	}
  else s = _first_r->Exec(s);

  // Store the success flag in a temporary storage.
  int success = s->ISuccess();

  // Restore the identifier environment.
  s->IOld();

  // s->Store() points to the transformed query. Make the new information
  // available in the identifier environment.

  if (_second) 
	s->IModifyFromHere(_second, 
			   s->Store(), 
			   matchNode->Which(), s->IEnv()->getData());


  // If the matching node was not the overall root of the tree, 
  // return to the old local root of the tree.
  if (!is_matchNode_new_one) s->SetStore(old_root);

  // Build it referring the current pattern in the last match (or match
  // each child) statement. The reconstruction will happen on the subtree
  // of local root of the tree.
  if (_second) 
	{
	  s->SetPatternDeep(_second,NULL);
	  s = updateDEP(_second, s, s->IEnv()->getData(), 1);
	  if (s->CurrentPattern())
		s = s->CurrentPattern()->buildCurrent(s);
	  else
		{
		  cout << "Error @ CRuleInvokeStmt : Exec" << endl;
		  cout << "Current pattern is NULL error " << endl;
		  exit(1);
		}
	}

  // Set it success value of rule invoke
  s->SetSuccess(success);
  
  if(_debug_)
	cout << "-- RuleInvokeStmt Ends" << endl;
  
  if (is_matchNode_new_one) 
	{ 
	  delete old_root; old_root = NULL; 
	  delete matchNode; matchNode = NULL; 
	}
  
  return s;
}

