/*
 * Copyright 1995, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *	Name:
 *	Organization:
 *	Address (postal and/or electronic):
 *
 * To:
 *	Software Librarian
 *	Computer Science Department, Box 1910
 *	Brown University
 *	Providence, RI 02912
 *
 *		or
 *
 *	brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.


//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc@cs.brown.edu)					 	    
//  FILE: COKO.l 
//  PROJ: COKO Compiler and Debugger
//  DATE: 4/96		                                                    
//
//  Lex file specifying tokens manipulated by Yacc for COKO compiler
//  and debugger
//
//////////////////////////////////////////////////////////////////////////////

*/

%{
#include "PAParser.h"
#include "Coko.tab.h"
#include <string.h>


int lineno = 1; 

%}


%%
[ \t]+                                                            {  }
[\n\f]                                                            {  lineno++; }
"-->"                                                             {  return GOESTO; }
[-][-][^>^\n]+                                                    {  } 
[+-]?[0-9]+                                                       {  CKlval.sval = strdup(yytext); 
                                                                    return INTCON; }
\"[^\"]*[\"]                                                  	  {  CKlval.sval = strdup(yytext); 
                                                                    return STRINGCON; }
[+-]?(([0-9]+)|([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?)                {  CKlval.sval = strdup(yytext);
                                                                    return FLOATCON; }

[Oo][Ii][Dd]                                                      {  return OID; }
"{"                                                               {  return ('{'); }  
"}"                                                               {  return ('}'); }     
"{|"                                                              {  return DPO; }   
"|}"                                                              {  return DPC; }   
[Ss][Tt][Rr]                                                      {  return KSTRING; }
[Rr][Ee][Aa][Ll]                                                  {  return KFLOAT; }
[Cc][Hh][Aa][Rr]                                                  {  return KCHAR;}

"="                                                               {  return PEQUAL; }
[Ee][Qq][Uu][Aa][Ll]                                              {  return PEQUAL; }  
  
"<"                                                               {  return LANGLE; }
">"                                                               {  return RANGLE; }
[Ll][Ee][Qq]                                                      {  return LE; } 
[Gg][Ee][Qq]                                                      {  return GE; }
[Ll][Tt]                                                          { return LESS; }
[Gg][Tt]                                                          { return GREATER; }
"<="                                                              {  return LE; } 
">="                                                              {  return GE; } 
"!="                                                              {  return NE; } 
","                                                               {  return (','); }  
"("                                                               {  return ('('); }  
")"                                                               {  return (')'); }  
"+"                                                               {  return PLUS; }  
"-"                                                               {  return MINUS; } 
"*"                                                               {  return MULT; }  
"/"                                                               {  return DIV; }  
"["                                                               {  return ('['); }  
"]"                                                               {  return (']'); }  
"&"                                                               {  return ('&'); }  
"|"                                                               {  return ('|'); }  
"\\/"           					                              {  return BOR; }  
"/\\"           					                              {  return BAND; }  
"~"            						                              {  return NOT; }  
[Pp][Rr][Ee][Dd]					                              {  return (PPREFIX); }
[Ff][Uu][Nn]						                              {  return (FPREFIX); }
[Oo][Bb][Jj]						                              {  return (OPREFIX); }
[Bb][Oo][Oo][Ll]					                              {  return (BPREFIX); }


"!"								                                  { return FINVOKE; }
"?"								                                  {  return PINVOKE; }
"o"					                                              {  return CIRC; }  
"K"                                                               {  return CONST; }  
"C"                                                               {  return CURRY; }  

[Ss][Hh][Ll]                					                  {  return SHIFTL; }  
[Ss][Hh][Rr]			                    		              {  return SHIFTR; }  
[Aa][Vv][Gg]							  {return FAVG; }
[Mm][Aa][Xx]                                                      {  return FMAX; }
[Mm][Ii][Nn]                                                      {  return FMIN; }
[Ss][Uu][Mm]                                                      {  return FSUM; }
[Cc][Oo][Uu][Nn][Tt]                                              {  return FCOUNT; }
[Ll][Ee][Xx]                                                      {  return PLEX; }
[Aa][Dd][Dd]                                                      {  return FADD; }
[Ss][Uu][Bb]                                                      {  return FSUB; }
[Mm][Uu][Ll]                                                      {  return FMUL; }
[Dd][Ii][Vv]                                                      {  return FDIV; }
[Nn][Ii][Ll]                                                      {  return NIL; }  
[Ll][Ee][Ss][Ss]                                                  {  return LESS; }  
[Gg][Rr][Ee][Aa][Tt][Ee][Rr]                                      {  return GREATER; }
[Ff][Aa][Ll][Ss][Ee]                                              {  return FALSEv; }  
[F]                                                               {  return FALSEv; }  
[Ii][Tt][Ee][Rr][Aa][Tt][Ee]                                      {  return ITERATE; }  
[Ii][Tt][Ee][Rr]                                                  {  return ITER; }  
[Pp][Cc][Oo][Nn]                                                  {  return PCON; }  
[Pp][Ff][Oo][Ll][Dd]                                              {  return PFOLD; }
[Ss][Ee][Tt]                                                      {  return SET; }  
[Ss][Ii][Nn][Gg][Ll][Ee][Tt][Oo][Nn]                              {  return SINGLETON; }
[Ff][Oo][Ll][Dd]                                                  {  return FOLD; }  
[Uu][Nn][Ii]                                                      {  return UNI; }  
[Ii][Nn][Tt]                                                      {  return INT; }  
[Ii][Ff]	                                                      {  return IF; }  
[Tt][Hh][Ee][Nn]                                                  {  return FTHEN; }  
[Ee][Ll][Ss][Ee]                                                  {  return ELSE; } 
[Ff][Ii]	                                                      {  return FI; }
[Dd][Ii][Ff]                                                      {  return DIF; }  
[Jj][Oo][Ii][Nn]                                                  {  return JOIN; }  
[Nn][Jj][Oo][Ii][Nn]                                              {  return NJOIN; }
[Uu][Nn][Nn][Ee][Ss][Tt]                                          {  return UNNEST; }  
[Ff][Ll][Aa][Tt]                                                  {  return FLATTEN; }  
[Cc][Hh][Oo][Oo][Ss][Ee]                                          {  return CHOOSE; }  
[Ee][Mm][Pp][Tt][Yy]                                              {  return EMPTY; }  
[Rr][Ss][Jj][Oo][Ii][Nn]                                          {  return RSJOIN; }
[Ll][Ss][Jj][Oo][Ii][Nn]                                          {  return LSJOIN; }
[Ii][Nn][Ss]                                                      {  return INS; }
[Ss][Oo][Rr][Tt]                                                  {  return SORT; }
[Aa][Nn][Dd]  	   	      	   	             	       	    	  {  return BAND; }  
[Oo][Rr]  	       	   	   	                  	       	    	  {  return BOR; }  
[Nn][Oo][Tt]  	      	   	     	        	    	          {  return NOT; }
[Ee][Xx][Ii][Ss][Tt][Ss]  	       	       	     	       	      {  return EXISTS; }  
[Ff][Oo][Rr][Aa][Ll][Ll]      	      	   	     	       	      {  return FORALL; }  
[Ii][Dd]                             		                      {  return ID; }  
[Tt][Ii][Dd]                             		                  {  return TID; }  
[#][1]                             				                  {  return ONE; }  
[#][2]                             				                  {  return TWO; }  
[Tt][Rr][Uu][Ee]                                                  {  return TRUEv; }  
[T]                                                               {  return TRUEv; }  
[Mm][Oo][Dd]				                         			  {  return MOD; }
[Aa][Bb][Ss]			                          				  {  return ABS; }
[Cc][Oo][Nn]           						                      {  return CON; } 
[Ee][Xx]            					                          {  return EX; }   
[Ff][Aa]	           					                          {  return FA; }

   
[Cc][Ii][Rr][Cc]						                          {  return CIRC; }
[Tt][Ii][Mm][Ee][Ss]						                      {  return TIMES; }  
[X] 								                              {  return XTIMES; }
\\[Ii][Nn][Vv]                                                    {  return FINV; }
[Ii][Nn][Vv]                                                      {  return INV; }
\\[Uu]                          				                      {  return UNION; }   
\\[Ii]                          				                      {  return INTER; }   
[Pp][Ll][Uu][Ss]                                                  {  return PLUS; }
[Oo][Pp][Ll][Uu][Ss]                                              {  return OPLUS; }      
[Ee][Xx][Cc][Ee][Pp][Tt]                                          {  return ODIF; }  

[Rr][Uu][Ll][Ee][Bb][Ll][Oo][Cc][Kk]                              {  return RBLOCK; }
[Tt][Rr][Aa][Nn][Ss][Ff][Oo][Rr][Mm][Aa][Tt][Ii][Oo][Nn]          {  return RBLOCK; }
"->"                                                              {  return THEN; }
[Pp][Rr][Ii][Nn][Tt]                                              {  return PRINT; }
[Pp][Rr][Ii][Nn][Tt][Oo][Qq][Ll]                                  {  return PRINTOQL; }
[Pp][Rr][Ii][Nn][Tt][Ss][Qq][Ll]                                  {  return PRINTSQL; }
[Nn][Ee][Qq]                                                      {  return NE; }  

[Ss][Kk][Ii][Pp]                                                  {  return SKIP; }  
[Uu][Ss][Ee][Ss]                                                  {  return USES; }  
[Mm][Aa][Tt][Cc][Hh]                                              {  return MATCH; }  
[Gg][Ii][Vv][Ee][Nn]                                              {  return GIVEN; }
[Dd][Oo]                                                          {  return DO; }
[Rr][Ee][Pp][Ee][Aa][Tt]                                          {  return REPEAT; }
[Tt][Dd]                                                          {  return TD; }
[Bb][Uu]                                                          {  return BU; }
[Bb][Ee][Gg][Ii][Nn]                                              {  return BEG; }  
[Ee][Nn][Dd]                                                      {  return END; }  
[Ee][Aa][Cc][Hh][Cc][Hh][Ii][Ll][Dd]                              {  return EACHCHILD; }  
[Cc][Uu][Rr][Rr][Tt][Rr][Ee][Ee]                                  {  return CURRTREE; }

":"                                                               {  return (':'); }  
";"                                                               {  return (';'); }  
"||"                                                              {  return OR; }

[_][Oo]            						                          {  return ODC; }  
[_][Ff]            						                          {  return FDC; }  
[_][Pp]            						                          {  return PDC; }  
[_][Bb]            						                          {  return BDC; }  
[_][Tt]            						                          {  return TDC; }  
[pqr][0-9']*                                                      {  CKlval.sval = strdup(yytext); 
                                                                    return PIDENT; }
(op)[0-9']*                                                       {  CKlval.sval = strdup(yytext); 
                                                                    return PIDENT; }
[fghjk][0-9']*                                                    {  CKlval.sval = strdup(yytext); 
                                                                    return FIDENT; }
[xyzABC][0-9']*                                                   {  CKlval.sval = strdup(yytext); 
                                                                    return OIDENT; }
[b][0-9']*                                                        {  CKlval.sval = strdup(yytext); 
                                                                    return BIDENT; }
[t][0-9']*                                                        {  CKlval.sval = strdup(yytext); 
                                                                    return TIDENT; }


[a-zA-Z][a-z_A-Z0-9]*[\*]*                  			          {  CKlval.sval = strdup(yytext); return IDENT; }
.                    						                      {return ERROR; }
%%



int yywrap ()  {return 1;}

