/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc)					 	    
//  FILE: ENEnvEntry.H 
//  PROJ: The "Karbonator"
//  DATE: 5/95		                                                    
//
// SYNOPSIS
//  Environment Entry Object.  Used to Pass Information about lambda bound
//   variables down tree during AQUA -> KOLA translation.
//  
//////////////////////////////////////////////////////////////////////////////

#ifndef KOLA_ENENVENTRY_HEADER
#define KOLA_ENENVENTRY_HEADER 

#include "../KGlobals.H"
#include "KOFunction.H"
#include "KOF1.H"
#include "KOF2.H"
#include "KOFID.H"
#include "KOFCompose.H"
#include "TYTObj.H"

#include <string.h>
#include <iostream.h>

class EnvEntry
{
    
 public:
  EnvEntry ();
  // Constructor
    
  EnvEntry (char *var, Function *f, TObj *t, EnvEntry* n = 0);
  // Constructor
    
  EnvEntry (EnvEntry& other);
  // Copy Constructor
    
  ~EnvEntry ();
  // Destructor
    
  char* show ();
  //  For debugging purposes.  Show this entry

  char *Variable () {return thevariable;}
  // Selector.
    
  Function* Funct () {return thefunction;}
  // Return function that is associated with this entry

  TObj *Type () {return thetype;}
  // Return type associated with this entry.
    
  EnvEntry* Next () {return next;}
  // What is the next environment entry in the table?
    
  void Next (EnvEntry *n) {next = n;}
  // Set the next environment entry in the table
    
  void Add_one ();
  // Modifies function of this entry by adding "o _1".  Then does the same
  // for next entry

  Function *GetFunc (char *var);
  // Return function that is associated with var found in this chain (left-most)
  // or NULL if none
    
  TObj *GetType (char *var);
  // Return type that is associated with var found in this chain (left-most)
  // or NULL if none
    
  EnvEntry *reversed ();
  // Return chain of copies of EnvEntries in chain buit in reversed order

  EnvEntry *append (EnvEntry *ent);
  // Append ent to end of environment
    
 protected:
   char *thevariable;
   // What is the variable stored in this entry

   Function *thefunction;
   // What is the function stored in this entry

   TObj *thetype;
   // If this entry is a label entry and the label refers
   // to a tuple, what is this tuple's label environment?
    
   EnvEntry *next;
   // What is the next entry in the table?
    
};


#endif


