/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOFTID.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOFTID.H"


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FTID
// Parameters:	(char* n)							  
//
//  Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FTID::FTID (TFunc * ftype) : Function (ftype, FUN_TID, "TID")
{
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FTID
// Parameters:	(FTID& other)							  
//
//  Copy Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FTID::FTID (const FTID& other) : Function ((Function &) other)
{
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~FTID
// Parameters:	()							  
//
//  Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FTID::~FTID ()
{
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy
// Parameters:	()							  
//
//  return copy of this as Function*
//         								    
//////////////////////////////////////////////////////////////////////////////

Function* FTID::Copy () const
{
  return new FTID (*this);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  deepeq
// Parameters:  ()                                                        
//
// Value-based equality.
// this is computed recursively.
// Return true if two trees have identical structures, otherwise return false.
// No unification occurs here.
//                                                                          
//////////////////////////////////////////////////////////////////////////////

int FTID::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == FNAME ) && (name_->bn == FUN_TID))
	{
	  delete name_;
	  return T;
	}
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  Myname
// Parameters:  ()
//
//  Returns a union type, NameUnion.
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, Index
//  and Bool, and those 5 different sub-classes use different enum type for
//  their own sub-classes' names, Myname methods needs a union type of
//  5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this
//  methods with NonType* to identify the specific instance of Kola object.
//
//////////////////////////////////////////////////////////////////////////////

NameUnion* FTID::Myname()
{
	NameUnion* a = new NameUnion(FNAME, FUN_TID);
	return a;
}

////////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FUSER: " + name
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FTID::representation ()
{
  char *res = new char [strlen ("TID") + 1];
  sprintf (res, "TID");
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* FTID::transform(CState* s)
{
    return new FTID(Type());
    
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
// For example, foo: g o g --> g o g is a legal rule.
//                  -------   -------
//                    |         |_\ 
//                    |_\         / building part
//                      / matching part 
//
// Assume that we apply this rule to a kola tree of the form "ID o ID."
// When unification occurs in a matching part, first we store first g and ID pair 
// into the env stack. Then the next time we face another g (second one), 
// we search the stack for the matching pair with variable name g and simply check 
// if the matching pair in the stack has same structure as the matching pair we have.
// In this case, second pair is (g, ID) as well and the rule gets fired successfully.
//
// Now assume that we apply the same rule to a kola tree of the form "ID o (ID o ID)."
// In this case, the rule fails since the second pair is (g, ID o ID),
// while the first pair is (g, ID). 
//
// Note that even in a case of failure, the rule itself is legal.
//
// "Duplication of a variable is legal in a rule section."
//
//////////////////////////////////////////////////////////////////////////////

int FTID::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();

  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_TID))
	{
	  delete tmp_name;
	  return T;
	}
  else
	{
	  delete tmp_name;
	  return F;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void FTID::replaceChild(NonType* child, int which_child)
{
}  

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* FTID::prolog ()
{
  char *res = new char [strlen ("ftid") + 1];
  sprintf (res, "ftid");
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref FTID::sp_prolog ()
{
  SP_term_ref res = SP_new_term_ref();
  SP_put_string(res,"ftid");
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	(koObject *o)
//
// Return the SQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* FTID::sql (koObject *o)
{    
  char *arg = o -> sql ();
  char *res = new char [strlen (".") + strlen (arg) + strlen ("tid") + 1];
  sprintf (res, "%s.tid", arg);
  delete [] arg;
  return res;
  	
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	(koObject *o)
//
// Return the OQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* FTID::oql (koObject *o)
{    
  char *arg = o -> oql ();
  char *res = new char [strlen (".") + strlen (arg) + strlen ("tid") + 1];
  sprintf (res, "%s.tid", arg);
  delete [] arg;
  return res;
  	
}

