/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KONonType.cc
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KONonType.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  unify
// Parameters:	(CState*)							  
//
//  Calls compares methods. If it succeed then return CState with true value.
//  Otherwise, return CState with false value.
//  This is for rule matching only.
//  Rule matching is different from statement matching.
//  Duplicated variable entry is allowed here. 
//  
//////////////////////////////////////////////////////////////////////////////

CState* NonType::unify(CState* a)
{
  
  if (!a->Store())
	{
	  cout << "EOR -------- EOR ---- Store is NULL :: NonType" << endl;
	  return NULL;
	}
  if (compares (a->Store(), a))
    a->SetSuccess(T);
  else
    a->SetSuccess(F);
  return a;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  unifySTM
// Parameters:	(CState*)							  
//
//  Calls comparesSTM methods. If it succeed then return CState with true value.
//  Otherwise, return CState with false value.
//  This method is for statement matching only.
//  No duplicated variable entry is allowed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* NonType::unifySTM(CState* a)
{
  
  if (!a->Store())
	{
	  cout << "EOR -------- EOR ---- Store is NULL :: NonType" << endl;
	  return NULL;
	}
  if (comparesSTM (a->Store(), a))
    a->SetSuccess(T);
  else
    a->SetSuccess(F);
  return a;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: isChildVar
//  Parameters:(NonType*)
//
//  Checks if NonType is a variable or not
//
//////////////////////////////////////////////////////////////////////////////

int NonType::isChildVar(NonType* nt)
{
  NameUnion* tmp_name = nt->Myname();

  if (!tmp_name)
	return F;

  if (((tmp_name->tag == BNAME) && (tmp_name->bn == BOOL_VAR)) 
	  || ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_VAR))
	  || ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_VAR))
	  || ((tmp_name->tag == ONAME) && (tmp_name->on == OBJ_VAR)))
	{
	  delete tmp_name;
	  return T;
	}
  else
	{
	  delete tmp_name;
	  return F;
	}
}

