/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOOPair.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOOPair.H"
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  OPair
// Parameters:	(koObject* o1, koObject *o2)		   	  
//
//  Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

OPair::OPair (koObject *o1, koObject *o2, TObj * t) : koObject (t,"[_ _]")
{
  one = o1;
  two = o2;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  OPair
// Parameters:	(OPair& other)							  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

OPair::OPair (const OPair& other) : koObject ((koObject &) other)
{
  one = (other . One ()) -> Copy ();
  two = (other . Two ()) -> Copy ();
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~OPair
// Parameters:	()							  
//
//  Destructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

OPair::~OPair ()
{
  delete one;
  delete two;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()							  
//
//  Virtual Copy Constructor 
//         								    
//////////////////////////////////////////////////////////////////////////////

koObject *OPair::Copy () const
{
  return new OPair (*this);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  deepeq
// Parameters:  ()
//
// Value-based equality.
// this is computed recursively.
// Return true if two trees have identical structures, otherwise return false.
// No unification occurs here.
//
//////////////////////////////////////////////////////////////////////////////
 
int OPair::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == ONAME ) && (name_->on == OBJ_PAIR))
  {
    delete name_;
    return ((one->deepeq(((OPair*)other)->One())) 
         && (two->deepeq(((OPair*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* OPair::Myname()
{
        NameUnion* a = new NameUnion(ONAME, OBJ_PAIR);
        return a;
}
 
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  nextChild
// Parameters:  ()
//
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* OPair::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	()							  
//  Return new OResPAIR 
//    (one -> reduce (), two -> reduce ())
//
/////////////////////////////////////////////////////////////////////////////

// OResult* OPair::reduce ()
// {
//   return new OResPAIR (one -> reduce (), two -> reduce ());
// }

// //////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FFO: "[" one -> representation () + ", " + two -> representation () + "]"
//         								    
//////////////////////////////////////////////////////////////////////////////

char* OPair::representation ()
{
  char *or = one -> representation ();
  char *tr = two -> representation ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("[, ]") + 1];
  sprintf (res, "[%s, %s]", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* OPair::transform(CState* s)
{
  koObject* first = (koObject*) One() -> transform(s);
  koObject* second = (koObject*) Two() -> transform(s);
  return new OPair(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  transformDeep
// Parameters:  (CState*, CIEnvStackDataType*)
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly.
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.)
//  Transform method searches the top frame only,
//  while TransformDeep methods searches entire stack starting from
//  a given point.
//
//////////////////////////////////////////////////////////////////////////////

NonType* OPair::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  koObject* first = (koObject*) One() -> transformDeep(s, startingPt);
  koObject* second = (koObject*) Two() -> transformDeep(s, startingPt);
  return new OPair(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here.
//
// For example, foo: g o g --> g o g is a legal rule.
//                  -------   -------
//                    |         |_\
//                    |_\         / building part
//                      / matching part
//
// Assume that we apply this rule to a kola tree of the form "ID o ID."
// When unification occurs in a matching part, first we store first g and ID pair
// into the env stack. Then the next time we face another g (second one),
// we search the stack for the matching pair with variable name g and simply check
// if the matching pair in the stack has same structure as the matching pair we have.
// In this case, second pair is (g, ID) as well and the rule gets fired successfully.
//
// Now assume that we apply the same rule to a kola tree of the form "ID o (ID o ID)."
// In this case, the rule fails since the second pair is (g, ID o ID),
// while the first pair is (g, ID).
//
// Note that even in a case of failure, the rule itself is legal.
//
// "Duplication of a variable is legal in a rule section."
//
//////////////////////////////////////////////////////////////////////////////

int OPair::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();

  if ((tmp_name->tag == ONAME) && (tmp_name->on == OBJ_PAIR))
	{
	  // do it for first child.
	  if(one->compares(((OPair*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one)) 
			// child (one) is a variable.
			{
			  if ((a->IFind(((OVar*)one)->Name())) == NULL)
				{ 
				  // child is a new variable.
				  a->IInsert(((OVar*)one)->Name(), new CIdent((OPair*)nt,((OPair*)nt)->One(),1));
				  // Push entry into the stack.  
				}
			  else 
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's 
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((OVar*)one)->Name()))->Child_Look_Up())->deepeq(((OPair*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++; 
	  // compare failed   
	 
	  // do it for second child.
	  if (two->compares(((OPair*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((OVar*)two)->Name())) == NULL)
				{ 
				  a->IInsert(((OVar*)two)->Name(), new CIdent((OPair*)nt, ((OPair*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((OVar*)two)->Name()))->Child_Look_Up())->deepeq(((OPair*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0. 
  // In this case, return false(0).
  // Otherwise, return true(1).

  delete tmp_name;

  if (return_val)
	return F;
  else
	return T;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for statement matching only.
// Statement matching is different from rule matching.
// No duplicated variable entry is allowed here.
//
// For example,
//             match f ! _O,
//                   f =  g o g  is a illegal statement (equation).
//                       -------
//                         |
//                         |_\
//                           / matching part
//
// Assume that we have a query tree of the form "ID o ID ! 3."
// Variable f gets bounded to "ID o ID" and we have to match "ID o ID" to "g o g."
// Matching "ID o ID" to "g o g" is perfectly legal in a rule section.
// However, it is illegal in statement section of the program.
// Remember that every variable on the right hand side of an equation must be a fresh one.
// However, after we match first g to ID, the second g is no longer fresh variable.
//
// "No duplication of a variable is legal in a statement section."
//
//////////////////////////////////////////////////////////////////////////////

int OPair::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();

  if ((tmp_name->tag == ONAME) && (tmp_name->on == OBJ_PAIR))
	{
	  // do it for first child.
	  if (one->comparesSTM(((OPair*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one)) 
			// child (one) is a variable.
			{
			  if ((a->IFind(((OVar*)one)->Name())) == NULL)
				{ 
				  // child is a new variable.
				  a->IInsert(((OVar*)one)->Name(), new CIdent((OPair*)nt,((OPair*)nt)->One(),1));
				  // Push entry into the stack.  
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((OVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++; 
	  // compare failed   
	 
	  // do it for second child.
	  if (two->comparesSTM(((OPair*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((OVar*)two)->Name())) == NULL)
				{ 
				  a->IInsert(((OVar*)two)->Name(), new CIdent((OPair*)nt, ((OPair*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((OVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0. 
  // In this case, return false(0).
  // Otherwise, return true(1).

  delete tmp_name;

  if (return_val)
	return F;
  else
	return T;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list (ancestor list) and inverse dependency list (descendant list)
//
//////////////////////////////////////////////////////////////////////////////

void OPair::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on descendant list (inverse dependency list).
// undo for dependency list (ancestor list) is not necessary, since
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void OPair::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void OPair::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* OPair::prolog ()
{
  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("opair(, )") + 1];
  sprintf (res, "opair(%s, %s)", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref OPair::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("opair"),2,or,tr);
  return res;
}
#endif // _PROLOG_


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	()
//
//  Return the SQL string equivalent of this object
//         								    
//////////////////////////////////////////////////////////////////////////////

char* OPair::sql ()
{
  char *arg1 = one -> sql ();
  char *arg2 = two -> sql ();

  if (strncmp (arg1, "ERROR", 5) == 0) {
	delete [] arg2;
	return arg1;
  } else if (strncmp (arg2, "ERROR", 5) == 0) {
	delete [] arg1;
	return arg2;
  } else {
	char *res = new char [strlen (" , ") + strlen (arg1) + strlen (arg2) + 1];
	sprintf (res, "%s, %s", arg1, arg2);
	delete [] arg1;
	delete [] arg2;
	return res;
  }

}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	()
//
//  Return the OQL string equivalent of this object
//         								    
//////////////////////////////////////////////////////////////////////////////

char* OPair::oql ()
{
  char *arg1 = one -> oql ();
  char *arg2 = two -> oql ();

  if (strncmp (arg1, "ERROR", 5) == 0) {
	delete [] arg2;
	return arg1;
  } else if (strncmp (arg2, "ERROR", 5) == 0) {
	delete [] arg1;
	return arg2;
  } else {
	char *res = new char [strlen ("struct (one: , two: )") + strlen (arg1) + strlen (arg2) + 1];
	sprintf (res, "struct (one: %s, two: %s)", arg1, arg2);
	delete [] arg1;
	delete [] arg2;
	return res;
  }

}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  asargs
// Parameters:	()
//
//  Return the OQL string equivalent of this object (used as collection of args
//  to function.
//
//////////////////////////////////////////////////////////////////////////////

char* OPair::asargs ()
{
  NameUnion *nu1 = one -> Myname ();
  NameUnion *nu2 = two -> Myname ();
  char *fst, *snd;

  if (nu1 -> on == OBJ_PAIR)
	fst = ((OPair *) one) -> asargs ();
  else
	fst = one -> oql ();

  if (nu2 -> on == OBJ_PAIR)
	snd = ((OPair *) two) -> asargs ();
  else
	snd = two -> oql ();

  char *res = new char [strlen (", ") + strlen (fst) + strlen (snd) + 1];
  sprintf (res, "%s, %s", fst, snd);
  delete [] fst;
  delete [] snd;
  delete nu1;
  delete nu2;
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  asargssql
// Parameters:	()
//
//  Return the SQL string equivalent of this object (used as collection of args
//  to function.
//
//////////////////////////////////////////////////////////////////////////////

char* OPair::asargssql (OPair *vlist)
{
  NameUnion *nu1 = one -> Myname ();
  NameUnion *nu2 = two -> Myname ();
  char *fst, *snd;

  if (nu1 -> on == OBJ_PAIR)
	fst = ((OPair *) one) -> asargssql (vlist -> One () -> asPair ());
  else {
	char *tmp1 = one -> sql ();
	char *tmp2 = vlist -> One () -> sql ();
	fst = new char [strlen (tmp1) + strlen (" ") + strlen (tmp2) + 1];
	sprintf (fst, "%s %s", tmp1, tmp2);
	delete [] tmp1;
	delete [] tmp2;
  };

  if (nu2 -> on == OBJ_PAIR)
	snd = ((OPair *) two) -> asargssql (vlist -> Two () -> asPair ());
  else {
	char *tmp1 = two -> sql ();
	char *tmp2 = vlist -> Two () -> sql ();
	snd = new char [strlen (tmp1) + strlen (" ") + strlen (tmp2) + 1];
	sprintf (snd, "%s %s", tmp1, tmp2);
	delete [] tmp1;
	delete [] tmp2;
  };

  char *res = new char [strlen (", ") + strlen (fst) + strlen (snd) + 1];
  sprintf (res, "%s, %s", fst, snd);
  delete [] fst;
  delete [] snd;
  delete nu1;
  delete nu2;
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  obj
// Parameters:	()
//
// Symbolic simplification of this.  Eventually,
// this method will replace "object" presently
// only defined with _OSTORE_ flag set
//  
//////////////////////////////////////////////////////////////////////////////

koObject* OPair::obj ()
{
  return new OPair (one -> obj (), two -> obj ());
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Vars
// Parameters:	()
//
// Return (potentially nested) pair of variables that 
// is homomorphic to this
//
//////////////////////////////////////////////////////////////////////////////

OPair* OPair::Vars () {

  NameUnion *nu1 = one -> Myname ();
  OPair *res;

  if (nu1 -> on == OBJ_PAIR) 
	res = new OPair (one -> asPair () -> Vars (), new OVar (0));
  else
	res = new OPair (new OVar (0), new OVar (0));

  delete nu1;
  return res;
}

