/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)					 	 
// FILE: KOPCElt.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////
 
#include "KOPCElt.H"
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PCElt
// Parameters:	(Predicate* p)							  
//
//  Constructor.  p is predicate argument.
//         								    
//////////////////////////////////////////////////////////////////////////////

PCElt::PCElt (Predicate* p, koObject *o, TPred * t) : Predicate (t, PRED_CELT,"C")
{
    one = p;
    two = o;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PCElt
// Parameters:	(PCElt& other)						   
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

PCElt::PCElt (const PCElt & other) : Predicate ((Predicate &) other)
{
    one = other . One () -> Copy ();
    two = other . Two () -> Copy ();
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~PCElt
// Parameters:	()
//
//  Destructor
//         								    
//////////////////////////////////////////////////////////////////////////////

PCElt::~PCElt ()
{
  delete one;
  delete two;  
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()
//
//  Return copy of *this as Predicate
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate* PCElt::Copy () const
{
    return new PCElt (*this);
}


//////////////////////////////////////////////////////////////////////////////
//
// Function Name: deepeq
// Parameters: (Nontype*)
//
// Value-based equality.
// this is computed recursively
//
//////////////////////////////////////////////////////////////////////////////

int PCElt::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == PNAME ) && (name_->pn == PRED_CELT))
  {
    delete name_;
    return ((one->deepeq(((PCElt*)other)->One())) && (two->deepeq(((PCElt*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* PCElt::Myname()
{
	NameUnion* a = new NameUnion(PNAME, PRED_CELT);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
//   Function Name: nextChild
//   Parameters: ()
// 
//  
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PCElt::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	(OResult* a)						      
//
//  Return
//    one -> reduce (
//      new OResPair (two -> reduce (), a)) 
//         								    
//////////////////////////////////////////////////////////////////////////////

// BResult* PCElt::reduce (OResult* a)
// {

//     BResult *res = 
//       (one -> reduce (new OResPAIR (two -> reduce (), a -> Copy ())));
//     delete a;
//     return res;
    
// }

// //////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "PFO: CELT" + one -> representation ()
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PCElt::representation ()
{

    char *pr = one -> representation ();
    char *or = two -> representation ();
    char *res = new char [strlen (pr) + strlen (or) + strlen ("C (, )") + 1];
    sprintf (res, "C (%s, %s)", pr, or);
    delete [] pr;
    delete [] or;    
    return res;

}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding Nontype object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* PCElt::transform(CState* s)
{
  Predicate* first = (Predicate*) One() -> transform(s);
  koObject* second = (koObject*) Two() -> transform(s);
  return new PCElt(first,second,Type());
    
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transformDeep
// Parameters:	(CState*, CIEnvStackDataType*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PCElt::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  Predicate* first = (Predicate*) One() -> transformDeep(s, startingPt);
  koObject* second = (koObject*) Two() -> transformDeep(s, startingPt);
  return new PCElt(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
//////////////////////////////////////////////////////////////////////////////

int PCElt::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
 
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_CELT))
	{
	  // do it for first child.
	  if (one->compares(((PCElt*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PCElt*)nt,((PCElt*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((PVar*)one)->Name()))->Child_Look_Up())->deepeq(((PCElt*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed  
	  
	  // do it for second child.
	  if (two->compares(((PCElt*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((OVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((OVar*)two)->Name(), new CIdent((PCElt*)nt, ((PCElt*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((OVar*)two)->Name()))->Child_Look_Up())->deepeq(((PCElt*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
  
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This match method is used for matching statement.
// No duplicated variable entry is allowed.
//
//////////////////////////////////////////////////////////////////////////////

int PCElt::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_CELT))
	{
	  // do it for first child.
	  if (one->comparesSTM(((PCElt*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PCElt*)nt,((PCElt*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed  
	  
	  // do it for second child.
	  if (two->comparesSTM(((PCElt*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((OVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((OVar*)two)->Name(), new CIdent((PCElt*)nt, ((PCElt*)nt)->Two(), 2));
				}
			  else
				{
				  return_val++;
				  cout << "Error : Variable " << ((OVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
  
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list and inverse dependency list
//
//////////////////////////////////////////////////////////////////////////////

void PCElt::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on inverse dependency list.
// undo for dependency list is not necessary, since 
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void PCElt::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void PCElt::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the currying predicate former.
//
//////////////////////////////////////////////////////////////////////////////

koBool* PCElt::invoke(koObject* x)
{
  return new BInvoke(this,x);
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the currying predicate former.
//
//////////////////////////////////////////////////////////////////////////////

OSBool* PCElt::exec(OSObject* x)
{
  return One()->exec(new OSPair(Two()->result(),x));
}
#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* PCElt::prolog ()
{
    char *pr = one -> prolog ();
    char *or = two -> prolog ();
    char *res = new char [strlen (pr) + strlen (or) + strlen ("pcelt(, )") + 1];
    sprintf (res, "pcelt(%s, %s)", pr, or);
    delete [] pr;
    delete [] or;    
    return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref PCElt::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("pcelt"),2,or,tr);
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  inv
// Parameters:	(koObject *o)
//
// Symbolic invocation of this on o.  Eventually,
// this method will replace "invoke" presently
// only defined with _OSTORE_ flag set
//
/////////////////////////////////////////////////////////////////////////////

koBool *PCElt::inv (koObject *o) 
{
  OPair *op = new OPair (two -> Copy (), o -> Copy ());
  koBool *res = one -> inv (op);
  delete op;
  return res;
}

